﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.TeamFoundation.WorkItemTracking.Client;
using Microsoft.TeamFoundation.Framework.Common;
using Microsoft.TeamFoundation.Framework.Server;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.Server;
using Microsoft.TeamFoundation.VersionControl.Server;
using System.Text.RegularExpressions;
using System.DirectoryServices.AccountManagement;

namespace CodePlex.TfsPluginSuite.Helper
{
    public sealed class TfsHelper : IDisposable
    {
        private TeamFoundationRequestContext requestContext;
        private CheckinNotification notification;
        private TfsTeamProjectCollection tfsTeamProjectCollection;

        public TfsHelper(TeamFoundationRequestContext requestContext, CheckinNotification notification)
        {
            this.requestContext = requestContext;
            this.notification = notification;
        }

        public NodeInfo GetIterationInfo(int workItemId)
        {
            var workItemStore = GetTeamProjectCollection().GetService<WorkItemStore>();
            var workItem = workItemStore.GetWorkItem(workItemId);
            Node iterationNode = workItemStore.Projects.GetById(workItem.Project.Id).FindNodeInSubTree(workItem.IterationId);
            var commonStructureService = (ICommonStructureService)GetTeamProjectCollection().GetService(typeof(ICommonStructureService));
            return commonStructureService.GetNode(iterationNode.Uri.ToString());
        }

        public int GetWorkItemQueryResultCount(string wiq)
        {
            var workItemStore = GetTeamProjectCollection().GetService<WorkItemStore>();
            var query = new Query(workItemStore, ResolveWiqMacros(wiq));
            return query.RunCountQuery();
        }

        public IEnumerable<PendingChange> GetPendingChanges()
        {
            var vcs = requestContext.GetService<TeamFoundationVersionControlService>();
            var submittedItems = notification.GetSubmittedItems(requestContext).Select(a => new ItemSpec(a, RecursionType.None)).ToArray();
            return vcs.QueryPendingChangesForWorkspace(requestContext, notification.WorkspaceName, notification.WorkspaceOwner.UniqueName, submittedItems, false, submittedItems.Length, null, true).CurrentEnumerable<PendingChange>();
        }

        private Uri GetTFSUri()
        {
            var locationService = this.requestContext.GetService<TeamFoundationLocationService>();
            return new Uri(locationService.GetServerAccessMapping(requestContext).AccessPoint + "/" + requestContext.ServiceHost.Name);
        }

        private TfsTeamProjectCollection GetTeamProjectCollection()
        {
            if (tfsTeamProjectCollection == null)
            {
                tfsTeamProjectCollection = new TfsTeamProjectCollection(GetTFSUri());
            }
            return tfsTeamProjectCollection;
        }

        private TeamFoundationIdentityService GetIdentityService()
        {
            return requestContext.GetService<TeamFoundationIdentityService>();
        }

        private string ResolveWiqMacros(string wiq)
        {
            var associatedWorkItems = notification.NotificationInfo.WorkItemInfo;
            string resolvedWiq = wiq;

            if (Regex.IsMatch(wiq, "@AssociatedWorkItems", RegexOptions.IgnoreCase))
            {
                var workItemIds = associatedWorkItems.AsEnumerable().Select(wiInfo => wiInfo.Id);
                var formattedWorkItemIds = string.Format("({0})", String.Join(",", workItemIds));
                resolvedWiq = Regex.Replace(resolvedWiq, "@AssociatedWorkItems", formattedWorkItemIds, RegexOptions.IgnoreCase);
            }

            if (Regex.IsMatch(wiq, "@Me", RegexOptions.IgnoreCase))
            {
                string userDisplayNameWithQuotes = "'" + GetUserDisplayName() + "'";
                resolvedWiq = Regex.Replace(resolvedWiq, "@Me", userDisplayNameWithQuotes, RegexOptions.IgnoreCase);
            }

            return resolvedWiq;
        }

        private string GetUserDisplayName()
        {
            var ims = requestContext.GetService<TeamFoundationIdentityService>();
            var user = ims.ReadIdentity(requestContext, IdentitySearchFactor.AccountName, requestContext.DomainUserName);
            return user.DisplayName;
        }

        public void Dispose()
        {
            if (tfsTeamProjectCollection != null)
            {
                tfsTeamProjectCollection.Dispose();
            }
        }
    }
}
